@@ -1,34 +1,27 @@
-# coding=utf-8
 """
 A protocol is an endpoint for EWS service connections. It contains all necessary information to make HTTPS connections.
 
 Protocols should be accessed through an Account, and are either created from a default Configuration or autodiscovered
 when creating an Account.
 """
-from __future__ import unicode_literals
-
 import datetime
 import logging
-from multiprocessing.pool import ThreadPool
 import os
 from threading import Lock
+from queue import LifoQueue, Empty, Full
 
-from cached_property import threaded_cached_property
 import requests.adapters
 import requests.sessions
 import requests.utils
-from future.utils import with_metaclass, python_2_unicode_compatible
-from future.moves.queue import LifoQueue, Empty, Full
-from oauthlib.oauth2 import BackendApplicationClient
+from oauthlib.oauth2 import BackendApplicationClient, WebApplicationClient
 from requests_oauthlib import OAuth2Session
-from six import string_types
 
-from .credentials import OAuth2Credentials
-from .errors import TransportError, SessionPoolMinSizeReached
+from .credentials import OAuth2AuthorizationCodeCredentials, OAuth2Credentials
+from .errors import TransportError, SessionPoolMinSizeReached, SessionPoolMaxSizeReached
 from .properties import FreeBusyViewOptions, MailboxData, TimeWindow, TimeZone
 from .services import GetServerTimeZones, GetRoomLists, GetRooms, ResolveNames, GetUserAvailability, \
     GetSearchableMailboxes, ExpandDL, ConvertId
-from .transport import get_auth_instance, get_service_authtype, NTLM, GSSAPI, SSPI, OAUTH2, DEFAULT_HEADERS
+from .transport import get_auth_instance, get_service_authtype, NTLM, OAUTH2, CREDENTIALS_REQUIRED, DEFAULT_HEADERS
 from .version import Version, API_VERSIONS
 
 log = logging.getLogger(__name__)
@@ -38,18 +31,22 @@ def close_connections():
     CachingProtocol.clear_cache()
 
 
-class BaseProtocol(object):
+class BaseProtocol:
     """Base class for Protocol which implements the bare essentials"""
 
     # The maximum number of sessions (== TCP connections, see below) we will open to this service endpoint. Keep this
     # low unless you have an agreement with the Exchange admin on the receiving end to hammer the server and
-    # rate-limiting policies have been disabled for the connecting user.
-    SESSION_POOLSIZE = 4
+    # rate-limiting policies have been disabled for the connecting user. Changing this setting only makes sense if
+    # you are using a thread pool to run multiple concurrent workers in this process.
+    SESSION_POOLSIZE = 1
     # We want only 1 TCP connection per Session object. We may have lots of different credentials hitting the server and
     # each credential needs its own session (NTLM auth will only send credentials once and then secure the connection,
-    # so a connection can only handle requests for one credential). Having multiple connections ser Session could
+    # so a connection can only handle requests for one credential). Having multiple connections per Session could
     # quickly exhaust the maximum number of concurrent connections the Exchange server allows from one client.
     CONNECTIONS_PER_SESSION = 1
+    # The number of times a session may be reused before creating a new session object. 'None' means "infinite".
+    # Discarding sessions after a certain number of usages may limit memory leaks in the Session object.
+    MAX_SESSION_USAGE_COUNT = 100
     # Timeout for HTTP requests
     TIMEOUT = 120
 
@@ -90,6 +87,15 @@ class BaseProtocol(object):
     def credentials(self):
         return self.config.credentials
 
+    @credentials.setter
+    def credentials(self, value):
+        # We are updating credentials, but that doesn't automatically propagate to the session objects. The simplest
+        # solution is to just kill the session pool and rebuild it.
+        with self._session_pool_lock:
+            self.config._credentials = value
+            self.close()
+            self._session_pool = self._create_session_pool()
+
     @property
     def retry_policy(self):
         return self.config.retry_policy
@@ -141,17 +147,9 @@ class BaseProtocol(object):
         # Autodetect and return authentication type
         raise NotImplementedError()
 
-    @classmethod
-    def get_useragent(cls):
-        if not cls.USERAGENT:
-            # import here to avoid a cyclic import
-            from exchangelib import __version__
-            cls.USERAGENT = "exchangelib/%s (%s)" % (__version__, requests.utils.default_user_agent())
-        return cls.USERAGENT
-
     def _create_session_pool(self):
         # Create a pool to reuse sessions containing connections to the server
-        session_pool = LifoQueue(maxsize=self._session_pool_size)
+        session_pool = LifoQueue()
         for _ in range(self._session_pool_size):
             session_pool.put(self.create_session(), block=False)
         return session_pool
@@ -160,9 +158,26 @@ class BaseProtocol(object):
     def session_pool_size(self):
         return self._session_pool_size
 
+    def increase_poolsize(self):
+        """Increases the session pool size. We increase by one session per call.
+        """
+        # Create a single session and insert it into the pool. We need to protect this with a lock while we are changing
+        # the pool size variable, to avoid race conditions. We must not exceed the SESSION_POOLSIZE limit.
+        if self._session_pool_size == self.SESSION_POOLSIZE:
+            raise SessionPoolMaxSizeReached('Session pool size cannot be increased further')
+        with self._session_pool_lock:
+            if self._session_pool_size >= self.SESSION_POOLSIZE:
+                log.debug('Session pool size was increased in another thread')
+                return
+            log.warning('Increasing session pool size from %s to %s', self._session_pool_size,
+                        self._session_pool_size + 1)
+            self._session_pool.put(self.create_session(), block=False)
+            self._session_pool_size += 1
+
     def decrease_poolsize(self):
         """Decreases the session pool size in response to error messages from the server requesting to rate-limit
         requests. We decrease by one session per call.
+
         """
         # Take a single session from the pool and discard it. We need to protect this with a lock while we are changing
         # the pool size variable, to avoid race conditions. We must keep at least one session in the pool.
@@ -172,7 +187,7 @@ class BaseProtocol(object):
             if self._session_pool_size <= 1:
                 log.debug('Session pool size was decreased in another thread')
                 return
-            log.warning('Lowering session pool size from %s to %s', self._session_pool_size,
+            log.warning('Decreasing session pool size from %s to %s', self._session_pool_size,
                         self._session_pool_size - 1)
             self.get_session().close()
             self._session_pool_size -= 1
@@ -184,6 +199,7 @@ class BaseProtocol(object):
                 log.debug('Server %s: Waiting for session', self.server)
                 session = self._session_pool.get(timeout=_timeout)
                 log.debug('Server %s: Got session %s', self.server, session.session_id)
+                session.usage_count += 1
                 return session
             except Empty:
                 # This is normal when we have many worker threads starving for available sessions
@@ -192,6 +208,9 @@ class BaseProtocol(object):
     def release_session(self, session):
         # This should never fail, as we don't have more sessions than the queue contains
         log.debug('Server %s: Releasing session %s', self.server, session.session_id)
+        if self.MAX_SESSION_USAGE_COUNT and session.usage_count > self.MAX_SESSION_USAGE_COUNT:
+            log.debug('Server %s: session %s usage exceeded limit. Discarding', self.server, session.session_id)
+            session = self.renew_session(session)
         try:
             self._session_pool.put(session, block=False)
         except Full:
@@ -211,48 +230,124 @@ class BaseProtocol(object):
         del session
         return self.create_session()
 
+    def refresh_credentials(self, session):
+        # Credentials need to be refreshed, probably due to an OAuth
+        # access token expiring. If we've gotten here, it's because the
+        # application didn't provide an OAuth client secret, so we can't
+        # handle token refreshing for it.
+        with self.credentials.lock:
+            if self.credentials.sig() == session.credentials_sig:
+                # Credentials have not been refreshed by another thread:
+                # they're the same as the session was created with. If
+                # this isn't the case, we can just go ahead with a new
+                # session using the already-updated credentials.
+                self.credentials.refresh()
+        return self.renew_session(session)
+
     def create_session(self):
         if self.auth_type is None:
             raise ValueError('Cannot create session without knowing the auth type')
-        if isinstance(self.credentials, OAuth2Credentials):
-            if self.auth_type != OAUTH2:
-                raise ValueError('Auth type must be %r for credentials type OAuth2Credentials' % OAUTH2)
-
-            token_url = 'https://login.microsoftonline.com/%s/oauth2/v2.0/token' % self.credentials.tenant_id
-            scope = ['https://outlook.office365.com/.default']
-            client = BackendApplicationClient(client_id=self.credentials.client_id)
-            session = self.raw_session(client)
-            # Fetch the token explicitly -- it doesn't occur implicitly
-            session.fetch_token(token_url=token_url, client_id=self.credentials.client_id,
-                                client_secret=self.credentials.client_secret, scope=scope)
-            session.auth = get_auth_instance(auth_type=OAUTH2, client=client)
-        elif self.credentials:
-            if self.auth_type == NTLM and self.credentials.type == self.credentials.EMAIL:
-                username = '\\' + self.credentials.username
-            else:
-                username = self.credentials.username
-            session = self.raw_session()
-            session.auth = get_auth_instance(auth_type=self.auth_type, username=username,
-                                             password=self.credentials.password)
-        else:
-            if self.auth_type not in (GSSAPI, SSPI):
+        if self.credentials is None:
+            if self.auth_type in CREDENTIALS_REQUIRED:
                 raise ValueError('Auth type %r requires credentials' % self.auth_type)
             session = self.raw_session()
             session.auth = get_auth_instance(auth_type=self.auth_type)
+        else:
+            with self.credentials.lock:
+                if isinstance(self.credentials, OAuth2Credentials):
+                    session = self.create_oauth2_session()
+                    # Keep track of the credentials used to create this session. If
+                    # and when we need to renew credentials (for example, refreshing
+                    # an OAuth access token), this lets us easily determine whether
+                    # the credentials have already been refreshed in another thread
+                    # by the time this session tries.
+                    session.credentials_sig = self.credentials.sig()
+                else:
+                    if self.auth_type == NTLM and self.credentials.type == self.credentials.EMAIL:
+                        username = '\\' + self.credentials.username
+                    else:
+                        username = self.credentials.username
+                    session = self.raw_session()
+                    session.auth = get_auth_instance(auth_type=self.auth_type, username=username,
+                                                     password=self.credentials.password)
+
         # Add some extra info
         session.session_id = sum(map(ord, str(os.urandom(100))))  # Used for debugging messages in services
+        session.usage_count = 0
         session.protocol = self
         log.debug('Server %s: Created session %s', self.server, session.session_id)
         return session
 
+    def create_oauth2_session(self):
+        if self.auth_type != OAUTH2:
+            raise ValueError('Auth type must be %r for credentials type OAuth2Credentials' % OAUTH2)
+
+        has_token = False
+        scope = ['https://outlook.office365.com/.default']
+        session_params = {}
+        token_params = {}
+
+        if isinstance(self.credentials, OAuth2AuthorizationCodeCredentials):
+            # Ask for a refresh token
+            scope.append('offline_access')
+
+            # We don't know (or need) the Microsoft tenant ID. Use
+            # common/ to let Microsoft select the appropriate tenant
+            # for the provided authorization code or refresh token.
+            #
+            # Suppress looks-like-password warning from Bandit.
+            token_url = 'https://login.microsoftonline.com/common/oauth2/v2.0/token'  # nosec
+
+            client_params = {}
+            has_token = self.credentials.access_token is not None
+            if has_token:
+                session_params['token'] = self.credentials.access_token
+            elif self.credentials.authorization_code is not None:
+                token_params['code'] = self.credentials.authorization_code
+                self.credentials.authorization_code = None
+
+            if self.credentials.client_id is not None and self.credentials.client_secret is not None:
+                # If we're given a client ID and secret, we have enough
+                # to refresh access tokens ourselves. In other cases the
+                # session will raise TokenExpiredError and we'll need to
+                # ask the calling application to refresh the token (that
+                # covers cases where the caller doesn't have access to
+                # the client secret but is working with a service that
+                # can provide it refreshed tokens on a limited basis).
+                session_params.update({
+                    'auto_refresh_kwargs': {
+                        'client_id': self.credentials.client_id,
+                        'client_secret': self.credentials.client_secret,
+                    },
+                    'auto_refresh_url': token_url,
+                    'token_updater': self.credentials.on_token_auto_refreshed,
+                })
+            client = WebApplicationClient(self.credentials.client_id, **client_params)
+        else:
+            token_url = 'https://login.microsoftonline.com/%s/oauth2/v2.0/token' % self.credentials.tenant_id
+            client = BackendApplicationClient(client_id=self.credentials.client_id)
+
+        session = self.raw_session(oauth2_client=client, oauth2_session_params=session_params)
+        if not has_token:
+            # Fetch the token explicitly -- it doesn't occur implicitly
+            token = session.fetch_token(token_url=token_url, client_id=self.credentials.client_id,
+                                        client_secret=self.credentials.client_secret, scope=scope,
+                                        **token_params)
+            # Allow the credentials object to update its copy of the new
+            # token, and give the application an opportunity to cache it
+            self.credentials.on_token_auto_refreshed(token)
+        session.auth = get_auth_instance(auth_type=OAUTH2, client=client)
+
+        return session
+
     @classmethod
-    def raw_session(cls, oauth2_client=None):
+    def raw_session(cls, oauth2_client=None, oauth2_session_params=None):
         if oauth2_client:
-            session = OAuth2Session(client=oauth2_client)
+            session = OAuth2Session(client=oauth2_client, **(oauth2_session_params or {}))
         else:
             session = requests.sessions.Session()
         session.headers.update(DEFAULT_HEADERS)
-        session.headers["User-Agent"] = cls.get_useragent()
+        session.headers['User-Agent'] = cls.USERAGENT
         session.mount('http://', adapter=cls.get_adapter())
         session.mount('https://', adapter=cls.get_adapter())
         return session
@@ -297,7 +392,7 @@ class CachingProtocol(type):
                 return protocol
             log.debug("Protocol __call__ cache miss. Adding key '%s'", str(_protocol_cache_key))
             try:
-                protocol = super(CachingProtocol, cls).__call__(*args, **kwargs)
+                protocol = super().__call__(*args, **kwargs)
             except TransportError as e:
                 # This can happen if, for example, autodiscover supplies us with a bogus EWS endpoint
                 log.warning('Failed to create cached protocol with key %s: %s', _protocol_cache_key, e)
@@ -317,76 +412,66 @@ class CachingProtocol(type):
         mcs._protocol_cache.clear()
 
 
-@python_2_unicode_compatible
-class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
+class Protocol(BaseProtocol, metaclass=CachingProtocol):
     def __init__(self, *args, **kwargs):
-        self.version_hint = None
-        super(Protocol, self).__init__(*args, **kwargs)
-
-        if not self.config.version:
-            # Version.guess() needs auth objects and a working session pool
-            self.config.version = Version.guess(self, hint=self.version_hint)
+        super().__init__(*args, **kwargs)
+        self._api_version_hint = None
+        self._version_lock = Lock()
 
     def get_auth_type(self):
         # Autodetect authentication type. We also set version hint here.
         name = str(self.credentials) if self.credentials and str(self.credentials) else 'DUMMY'
-        auth_type, version_hint = get_service_authtype(
-            service_endpoint=self.service_endpoint, versions=API_VERSIONS, name=name
+        auth_type, api_version_hint = get_service_authtype(
+            service_endpoint=self.service_endpoint, retry_policy=self.retry_policy, api_versions=API_VERSIONS, name=name
         )
-        self.version_hint = version_hint
+        self._api_version_hint = api_version_hint
         return auth_type
 
     @property
     def version(self):
+        # Make sure only one thread does the guessing.
+        if not self.config.version or not self.config.version.build:
+            with self._version_lock:
+                if not self.config.version or not self.config.version.build:
+                    # Version.guess() needs auth objects and a working session pool
+                    self.config.version = Version.guess(self, api_version_hint=self._api_version_hint)
         return self.config.version
 
-    @threaded_cached_property
-    def thread_pool(self):
-        # Used by services to process service requests that are able to run in parallel. Thread pool should be
-        # larger than the connection pool so we have time to process data without idling the connection.
-        # Create the pool as the last thing here, since we may fail in the version or auth type guessing, which would
-        # leave open threads around to be garbage collected.
-        thread_poolsize = 4 * self._session_pool_size
-        return ThreadPool(processes=thread_poolsize)
-
-    def close(self):
-        log.debug('Server %s: Closing thread pool', self.server)
-        # Close the thread pool before closing the session pool to ensure all sessions are released.
-        if "thread_pool" in self.__dict__:
-            # Calling thread_pool.join() in Python 3.8 will hang forever. This is seen when running a test case that
-            # uses the thread pool, e.g.: python tests/__init__.py MessagesTest.test_export_with_error
-            # I don't know yet why this is happening.
-            self.thread_pool.terminate()
-            del self.__dict__["thread_pool"]
-        super(Protocol, self).close()
-
     def get_timezones(self, timezones=None, return_full_timezone_data=False):
-        """ Get timezone definitions from the server
+        """Get timezone definitions from the server
+
+        Args:
+          timezones: A list of EWSDateTime instances. If None, fetches all timezones from server (Default value = None)
+          return_full_timezone_data: If true, also returns periods and transitions (Default value = False)
+
+        Returns:
+          A list of (tz_id, name, periods, transitions) tuples
 
-        :param timezones: A list of EWSDateTime instances. If None, fetches all timezones from server
-        :param return_full_timezone_data: If true, also returns periods and transitions
-        :return: A list of (tz_id, name, periods, transitions) tuples
         """
         return GetServerTimeZones(protocol=self).call(
             timezones=timezones, return_full_timezone_data=return_full_timezone_data
         )
 
     def get_free_busy_info(self, accounts, start, end, merged_free_busy_interval=30, requested_view='DetailedMerged'):
-        """ Returns free/busy information for a list of accounts
+        """Returns free/busy information for a list of accounts
+
+        Args:
+          accounts: A list of (account, attendee_type, exclude_conflicts) tuples, where account is either an Account
+            object or a string, attendee_type is a MailboxData.attendee_type choice, and exclude_conflicts is a boolean.
+          start: The start datetime of the request
+          end: The end datetime of the request
+          merged_free_busy_interval: The interval, in minutes, of merged free/busy information (Default value = 30)
+          requested_view: The type of information returned. Possible values are defined in the
+            FreeBusyViewOptions.requested_view choices. (Default value = 'DetailedMerged')
+
+        Returns:
+          A generator of FreeBusyView objects
 
-        :param accounts: A list of (account, attendee_type, exclude_conflicts) tuples, where account is an Account
-               object, attendee_type is a MailboxData.attendee_type choice, and exclude_conflicts is a boolean.
-        :param start: The start datetime of the request
-        :param end: The end datetime of the request
-        :param merged_free_busy_interval: The interval, in minutes, of merged free/busy information
-        :param requested_view: The type of information returned. Possible values are defined in the
-               FreeBusyViewOptions.requested_view choices.
-        :return: A generator of FreeBusyView objects
         """
         from .account import Account
         for account, attendee_type, exclude_conflicts in accounts:
-            if not isinstance(account, Account):
-                raise ValueError("'accounts' item %r must be an 'Account' instance" % account)
+            if not isinstance(account, (Account, str)):
+                raise ValueError("'accounts' item %r must be an 'Account' or 'str' instance" % account)
             if attendee_type not in MailboxData.ATTENDEE_TYPES:
                 raise ValueError("'accounts' item %r must be one of %s" % (attendee_type, MailboxData.ATTENDEE_TYPES))
             if not isinstance(exclude_conflicts, bool):
@@ -410,7 +495,7 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
                     for_year=start.year
                 ),
                 mailbox_data=[MailboxData(
-                    email=account.primary_smtp_address,
+                    email=account.primary_smtp_address if isinstance(account, Account) else account,
                     attendee_type=attendee_type,
                     exclude_conflicts=exclude_conflicts
                 ) for account, attendee_type, exclude_conflicts in accounts],
@@ -429,34 +514,36 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
         return GetRooms(protocol=self).call(roomlist=RoomList(email_address=roomlist))
 
     def resolve_names(self, names, return_full_contact_data=False, search_scope=None, shape=None):
-        """ Resolve accounts on the server using partial account data, e.g. an email address or initials
+        """Resolve accounts on the server using partial account data, e.g. an email address or initials
+
+        Args:
+          names: A list of identifiers to query
+          return_full_contact_data: If True, returns full contact data (Default value = False)
+          search_scope: The scope to perform the search. Must be one of SEARCH_SCOPE_CHOICES (Default value = None)
+          shape: return: A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact)
+            items (Default value = None)
+
+        Returns:
+          A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact) items
 
-        :param names: A list of identifiers to query
-        :param return_full_contact_data: If True, returns full contact data
-        :param search_scope: The scope to perform the search. Must be one of SEARCH_SCOPE_CHOICES
-        :param shape:
-        :return: A list of Mailbox items or, if return_full_contact_data is True, tuples of (Mailbox, Contact) items
         """
-        from .items import SHAPE_CHOICES, SEARCH_SCOPE_CHOICES
-        if search_scope:
-            if search_scope not in SEARCH_SCOPE_CHOICES:
-                raise ValueError("'search_scope' %s must be one if %s" % (search_scope, SEARCH_SCOPE_CHOICES))
-        if shape:
-            if shape not in SHAPE_CHOICES:
-                raise ValueError("'shape' %s must be one if %s" % (shape, SHAPE_CHOICES))
         return list(ResolveNames(protocol=self).call(
             unresolved_entries=names, return_full_contact_data=return_full_contact_data, search_scope=search_scope,
             contact_data_shape=shape,
         ))
 
     def expand_dl(self, distribution_list):
-        """ Expand distribution list into it's members
+        """Expand distribution list into it's members
+
+        Args:
+          distribution_list: SMTP address of the distribution list to expand, or a DLMailbox representing the list
+
+        Returns:
+          List of Mailbox items that are members of the distribution list
 
-        :param distribution_list: SMTP address of the distribution list to expand, or a DLMailbox representing the list
-        :return: List of Mailbox items that are members of the distribution list
         """
         from .properties import DLMailbox
-        if isinstance(distribution_list, string_types):
+        if isinstance(distribution_list, str):
             distribution_list = DLMailbox(email_address=distribution_list, mailbox_type='PublicDL')
         return list(ExpandDL(protocol=self).call(distribution_list=distribution_list))
 
@@ -464,9 +551,13 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
         """This method is only available to users who have been assigned the Discovery Management RBAC role. See
         https://docs.microsoft.com/en-us/exchange/permissions-exo/permissions-exo
 
-        :param search_filter: Is set, must be a single email alias
-        :param expand_group_membership: If True, returned distribution lists are expanded
-        :return: a list of SearchableMailbox, FailedMailbox or Exception instances
+        Args:
+          search_filter: Is set, must be a single email alias (Default value = None)
+          expand_group_membership: If True, returned distribution lists are expanded (Default value = False)
+
+        Returns:
+          a list of SearchableMailbox, FailedMailbox or Exception instances
+
         """
         return list(GetSearchableMailboxes(protocol=self).call(
             search_filter=search_filter,
@@ -476,13 +567,15 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
     def convert_ids(self, ids, destination_format):
         """Converts item and folder IDs between multiple formats
 
-        :param ids: a list of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
-        :param destination_format: A string
-        :return: a generator of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
+        Args:
+          ids: a list of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
+          destination_format: A string
+
+        Returns:
+          a generator of AlternateId, AlternatePublicFolderId or AlternatePublicFolderItemId instances
+
         """
-        from .properties import ID_FORMATS, AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
-        if destination_format not in ID_FORMATS:
-            raise ValueError("'destination_format' %r must be one of %s" % (destination_format, ID_FORMATS))
+        from .properties import AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
         cls_map = {cls.response_tag(): cls for cls in (
             AlternateId, AlternatePublicFolderId, AlternatePublicFolderItemId
         )}
@@ -494,8 +587,9 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
                 yield id_cls.from_xml(i, account=None)
 
     def __getstate__(self):
-        # The thread and session pools cannot be pickled
-        state = super(Protocol, self).__getstate__()
+        # The lock and thread pool cannot be pickled
+        state = super().__getstate__()
+        del state['_version_lock']
         try:
             del state['thread_pool']
         except KeyError:
@@ -503,30 +597,45 @@ class Protocol(with_metaclass(CachingProtocol, BaseProtocol)):
             pass
         return state
 
+    def __setstate__(self, state):
+        # Restore the lock. The thread pool is a cached property and will be recreated automatically.
+        self.__dict__.update(state)
+        self._version_lock = Lock()
+
     def __str__(self):
+        # Don't trigger version guessing here just for the sake of printing
+        if self.config.version:
+            fullname, api_version, build = self.version.fullname, self.version.api_version, self.version.build
+        else:
+            fullname, api_version, build = '[unknown]', '[unknown]', '[unknown]'
+
         return '''\
 EWS url: %s
 Product name: %s
 EWS API version: %s
 Build number: %s
-EWS auth: %s''' % (
-            self.service_endpoint,
-            self.version.fullname,
-            self.version.api_version,
-            self.version.build,
-            self.auth_type,
-        )
+EWS auth: %s''' % (self.service_endpoint, fullname, api_version, build, self.auth_type)
 
 
 class NoVerifyHTTPAdapter(requests.adapters.HTTPAdapter):
     """An HTTP adapter that ignores TLS validation errors. Use at own risk."""
+
     def cert_verify(self, conn, url, verify, cert):
         # pylint: disable=unused-argument
         # We're overiding a method so we have to keep the signature
-        super(NoVerifyHTTPAdapter, self).cert_verify(conn=conn, url=url, verify=False, cert=cert)
+        super().cert_verify(conn=conn, url=url, verify=False, cert=cert)
 
 
-class RetryPolicy(object):
+class TLSClientAuth(requests.adapters.HTTPAdapter):
+    """An HTTP adapter that implements Certificate Based Authentication (CBA)"""
+    cert_file = None
+
+    def init_poolmanager(self, *args, **kwargs):
+        kwargs['cert_file'] = self.cert_file
+        return super().init_poolmanager(*args, **kwargs)
+
+
+class RetryPolicy:
     """Stores retry logic used when faced with errors from the server"""
     @property
     def fail_fast(self):
@@ -542,6 +651,9 @@ class RetryPolicy(object):
     def back_off_until(self, value):
         raise NotImplementedError()
 
+    def back_off(self, seconds):
+        raise NotImplementedError()
+
 
 class FailFast(RetryPolicy):
     """Fail immediately on server errors"""
@@ -557,7 +669,12 @@ class FailFast(RetryPolicy):
 class FaultTolerance(RetryPolicy):
     """Enables fault-tolerant error handling. Tells internal methods to do an exponential back off when requests start
     failing, and wait up to max_wait seconds before failing.
+
     """
+
+    # Back off 60 seconds if we didn't get an explicit suggested value
+    DEFAULT_BACKOFF = 60
+
     def __init__(self, max_wait=3600):
         self.max_wait = max_wait
         self._back_off_until = None
@@ -598,7 +715,7 @@ class FaultTolerance(RetryPolicy):
 
     def back_off(self, seconds):
         if seconds is None:
-            seconds = 60  # Back off 60 seconds if we didn't get an explicit suggested value
+            seconds = self.DEFAULT_BACKOFF
         value = datetime.datetime.now() + datetime.timedelta(seconds=seconds)
         with self._back_off_lock:
             self._back_off_until = value
